/** 
  Second Life/OpenSim Animation Relay Server 1.0.0 	

						anm_server.c v1.0  by Fumi.Iseki (C)2011

*/



#include "ipaddr_tool.h"

#include "anm_server.h"
#include "anm_data.h"




void   anm_server(int usock, char* idkey)
{
	struct sockaddr_in rcv_addr;
	struct sockaddr_in snd_addr;

	char*  ipaddr     = NULL;
    char*  ipaddrnum  = NULL;
	udp_header* dat   = NULL;

	struct 	sigaction sa;
	time_t  ntime, stime;


	DEBUG_MODE print_message("[%d] ANM_SERVER: Animation Relay Server start.\n", CrntPID);

	// シグナルハンドリング
	sa.sa_handler = anm_server_term;
	sa.sa_flags   = 0;
	sigemptyset(&sa.sa_mask);
	sigaddset(&sa.sa_mask, SIGCHLD);
	sigaction(SIGINT,  &sa, NULL);
	sigaction(SIGHUP,  &sa, NULL);
	sigaction(SIGTERM, &sa, NULL);

	init_rand();

	stime = time(NULL);
	Buffer buf = make_Buffer(LFRAME);
    tList* Relay_List = add_tList_node_bystr(NULL, 0, 0, LIST_ANCHOR, NULL, NULL, 0);
	
	int ssock = socket(AF_INET, SOCK_DGRAM, 0);		// 送信用ソケット


	Loop {
		ntime = time(NULL);

		if (ntime-stime>MaxIdleTime) {
			DEBUG_MODE print_message("[%d] ANM_SERVER: long idle time. go to die.\n", CrntPID);
			break;
		}


		////////////////////////////////////////////////////////////////////////////////////////////////////////
		// Data受信

		if (recv_wait(usock, ANM_UDP_TIME_OUT)) {
			int cc = udp_recv_Buffer(usock, &buf, &rcv_addr);
			//DEBUG_MODE print_message("[%d] ANM_SERVER: WARNING: RECEIVE = %d\n", CrntPID, cc);
			if (cc<(int)sizeof(udp_header)) continue;

			ipaddr    = get_ipaddr(rcv_addr.sin_addr);
            ipaddrnum = (char*)to_address_num4(ipaddr, 0);

			if (Allow_IPaddr!=NULL && !is_host_in_list(Allow_IPaddr, (unsigned char*)ipaddrnum, NULL)) {
				DEBUG_MODE print_message("[%d] ANM_SERVER: WARNING: udp data from unknown node!! [%s]\n", CrntPID, ipaddr);
				freeNull(ipaddr);
				freeNull(ipaddrnum);
				continue;
			}

			dat = (udp_header*)buf.buf;
			if (strncmp(idkey, dat->key, ANM_COM_LEN_IDKEY)) {
				DEBUG_MODE print_message("[%d] ANM_SERVER: WARNING: not match IDKEY of received data!! [%s]\n", CrntPID, ipaddr);
				freeNull(ipaddr);
				freeNull(ipaddrnum);
				continue;
			}
		}
		else continue;



		////////////////////////////////////////////////////////////////////////////////////////////////////////
		// コマンド処理

		unsigned short port = ntohs(dat->port);
		tList* lt = search_regist(Relay_List->next, (int)port, ipaddr);

		//
		if (dat->com[0]==ANM_COM_REQ_TRANSFER)
		{
			if (lt!=NULL && !strncasecmp(dat->uuid, lt->ldat.val.buf, ANM_COM_LEN_UUID)) {
				tList* pp = Relay_List->next;
				while (pp!=NULL) {
					snd_addr = get_sockaddr((char*)pp->ldat.key.buf, (unsigned short)pp->ldat.lv);
					//DEBUG_MODE print_message("[%d] ANM_SERVER: send data to %s:%d (%d)\n", CrntPID, pp->ldat.key.buf, pp->ldat.lv, ntohs(dat->sz));
					udp_send_Buffer(ssock, &buf, &snd_addr);
					pp = pp->next;
				}
			}
		}

		//
		else if (dat->com[0]==ANM_COM_REQ_REGIST)
		{
			if (lt==NULL) {
				add_tList_node_bystr(Relay_List, 0, (int)port, ipaddr, dat->uuid, NULL, 0);
				Relay_List->ldat.id++;

				DEBUG_MODE {
					print_message("[%d] ANM_SERVER: Regist Request from %s:%d\n", CrntPID, ipaddr, port);
                	print_message("==============================================\n");
                	print_tList(stderr, Relay_List);
                	print_message("==============================================\n");
				}
				ANM_LOG_FILE print_log(LogFile, "%s [%d][%s] Regist Request from [%s:%d]. Registered.\n", get_local_time(), CrntPID, idkey, ipaddr, port);
			}
			else {
				DEBUG_MODE  print_message("[%d] ANM_SERVER: Already Registered  %s:%d\n", CrntPID, ipaddr, port);
				ANM_LOG_FILE print_log(LogFile, "%s [%d][%s] Regist Request from [%s:%d]. Already Registered.\n", get_local_time(), CrntPID, idkey, ipaddr, port);
			}
		}

		//
		else if (dat->com[0]==ANM_COM_REQ_DELETE)
		{
			if (lt!=NULL && !strncasecmp(dat->uuid, lt->ldat.val.buf, ANM_COM_LEN_UUID)) {
				tList* pp = Relay_List->next;
				while (pp!=NULL) {
					snd_addr = get_sockaddr((char*)pp->ldat.key.buf, (unsigned short)pp->ldat.lv);
					DEBUG_MODE print_message("[%d] ANM_SERVER: Send Delete data to %s:%d\n", CrntPID, pp->ldat.key.buf, pp->ldat.lv);
					udp_send_Buffer(ssock, &buf, &snd_addr);
					pp = pp->next;
				}
				DEBUG_MODE print_message("[%d] ANM_SERVER: Delete Requestfrom  %s:%d\n", CrntPID, ipaddr, port);
				ANM_LOG_FILE print_log(LogFile, "%s [%d][%s] Delete Request from [%s:%d]. Animation UUID is [%s].\n", get_local_time(), CrntPID, idkey, ipaddr, port, dat->uuid);
			}
		}

		//
		else if (dat->com[0]==ANM_COM_REQ_LOGOUT)
		{
			if (lt!=NULL && !strncasecmp(dat->uuid, lt->ldat.val.buf, ANM_COM_LEN_UUID)) {
				del_tList_node(lt);
				Relay_List->ldat.id--;
			}

			DEBUG_MODE {
				print_message("[%d] ANM_SERVER: Logout Request from %s:%d\n", CrntPID, ipaddr, port);
                print_message("==============================================\n");
                print_tList(stderr, Relay_List);
                print_message("==============================================\n");
			}
			ANM_LOG_FILE print_log(LogFile, "%s [%d][%s] Logout Request from [%s:%d].\n", get_local_time(), CrntPID, idkey, ipaddr, port);
		}

		freeNull(ipaddr);
		freeNull(ipaddrnum);

		stime = ntime;
	}


	//////////////////////////////////////////////////////////////////////////////////////////////////////////
	//

	socket_close(ssock);
	del_all_tList(&Relay_List);				


	DEBUG_MODE print_message("[%d] ANM_SERVER: down!!\n", CrntPID);
	ANM_LOG_FILE print_log(LogFile, "%s [%d][%s] Relay Process is going down.\n", get_local_time(), CrntPID, idkey);
	return;
}





///////////////////////////////////////////////////////////////////////////////////
//
// tList
//


/**
void del_process_list(tList* pl, int pid)

	機能：
		プロセスIDをキーにして，リストから削除

*/
void del_process_list(tList* pl, int pid)
{
	if (pl==NULL) return;
	tList* pp = pl;

	DEBUG_MODE print_message("[%d] DEL_PROCESS_LIST: called with %d\n", CrntPID, pid);
	while (pp!=NULL) {
		if (pp->ldat.id==(int)pid) {
			del_tList_node(pp);
			pl->ldat.id--;				// 項目の数
			break;
		}
		pp = pp->next;
	}

	return; 
}




tList*  search_regist(tList* lp, int port, char* ip)
{
	while(lp!=NULL) {

		if (!strcmp(ip, lp->ldat.key.buf)) {
			if (port==lp->ldat.lv) return lp;
		}
		lp = lp->next;	
	}

	return NULL;
}








//////////////////////////////////////////////////////////////////////////////////////////////////////////
// Termination of Program
//

void  anm_server_term(int sig)
{
	exit(0);
}



